#include <cv.h>
#include <cvaux.h>
#include <highgui.h>
#include <omp.h>

#define _E_ 1.0f

struct hsImage{
	int width, height;
	int nChannels; //Pixel color channel
	int numPixels;
	float* iPixel;
	int* constraint;
	float* alphaMap;
	hsImage* parent;
	bool hasF, hasB, hasU;
	int uX0, uY0, uX1, uY1;
	float* vX; //vector for x
	float* vP; //vector for y
	float* vR; //vector for z

	//Optional
	float** meanPixel;
	float** covarianceMatrix;
	float** AkStar;
	float* BkStar;

	hsImage(uchar* sourceImage, uchar* sourceAlphamap, int sWidth, int sHeight, int sourceChannels)
	{
		iPixel = new float[sWidth*sHeight*sourceChannels];
		constraint = new int[sWidth*sHeight];
		alphaMap = new float[sWidth*sHeight];
		vX = new float[sWidth*sHeight];
		vP = new float[sWidth*sHeight];
		vR = new float[sWidth*sHeight];

		width = sWidth;
		height = sHeight;
		nChannels = sourceChannels;
		numPixels = width * height;
		parent = NULL;
		hasF = false;	hasU = false;	hasB = false;
		uX0 = width+1; uY0 = height+1; uX1 = -1; uY1 = -1;
		meanPixel = NULL;	covarianceMatrix = NULL;
		AkStar = NULL;	BkStar = NULL;

		const int isForeground = 255 * nChannels;
		const int isBackground = 0;
		for(int i=0; i<numPixels; i++)
		{
			//set constraint map and alpha map
			int state = 0;
			for(int c=0; c<nChannels; c++)
				state += sourceAlphamap[i*nChannels + c];
			if(state == isForeground)
			{
				constraint[i] = 1;
				alphaMap[i] = 1.0f;
			}
			else if(state == isBackground)
			{
				constraint[i] = 1;
				alphaMap[i] = 0.0f;
			}
			else
			{
				constraint[i] = 0;
				alphaMap[i] = 0.5f;
			}
			//set pixel color
			for(int c=0; c<nChannels; c++)
				iPixel[i*nChannels + c] = float(sourceImage[i*nChannels + c])/255.0f;
		}

		checkFBU();
	}

	hsImage(hsImage* sourceImage, int x0, int y0, int x1, int y1)
	{
		width = x1 - x0;
		height = y1 - y0;
		nChannels = sourceImage->nChannels;
		numPixels = width * height;
		parent = sourceImage;
		hasF = false;	hasU = false;	hasB = false;
		uX0 = width+1; uY0 = height+1; uX1 = -1; uY1 = -1;
		meanPixel = NULL;	covarianceMatrix = NULL;
		AkStar = NULL;	BkStar = NULL;

		iPixel = new float[width*height*nChannels];
		constraint = new int[width*height];
		alphaMap = new float[width*height];
		vX = new float[width*height];
		vP = new float[width*height];
		vR = new float[width*height];

		for(int y=y0; y<y1; y++){
			for(int x=x0; x<x1; x++)
			{
				int sX = x-x0;
				int sY = y-y0;
				for(int c=0; c<nChannels; c++)
					iPixel[sY * (width*nChannels) + sX*nChannels + c] =
					sourceImage->iPixel[y*(sourceImage->width*nChannels)+x*nChannels+c];
				constraint[sY*width + sX] = sourceImage->constraint[y*(sourceImage->width) + x];
				alphaMap[sY*width + sX] = sourceImage->alphaMap[y*(sourceImage->width) + x];
				vX[sY*width + sX] = sourceImage->vX[y*(sourceImage->width) + x];
				vP[sY*width + sX] = sourceImage->vP[y*(sourceImage->width) + x];
				vR[sY*width + sX] = sourceImage->vR[y*(sourceImage->width) + x];
			}
		}
		checkFBU();
	}

	hsImage(hsImage* sourceImage)
	{
		width = sourceImage->width;
		height = sourceImage->height;
		nChannels = sourceImage->nChannels;
		numPixels = sourceImage->numPixels;
		parent = sourceImage->parent;
		hasF = sourceImage->hasF;	hasU = sourceImage->hasU;	hasB = sourceImage->hasB;
		uX0 = sourceImage->uX0; uY0 = sourceImage->uY0; uX1 = sourceImage->uX1; uY1 = sourceImage->uY1;
		AkStar = NULL;	BkStar = NULL;
		meanPixel = NULL;	covarianceMatrix = NULL;
		

		iPixel = new float[width*height*nChannels];
		constraint = new int[width*height];
		alphaMap = new float[width*height];
		vX = new float[width*height];
		vP = new float[width*height];
		vR = new float[width*height];

		for(int i=0; i<numPixels; i++)
		{
			for(int c=0; c<nChannels; c++)
				iPixel[i*nChannels + c] = sourceImage->iPixel[i*nChannels + c];
			constraint[i] = sourceImage->constraint[i];
			alphaMap[i] = sourceImage->alphaMap[i];
			vX[i] = sourceImage->vX[i];
			vP[i] = sourceImage->vP[i];
			vR[i] = sourceImage->vR[i];
		}
	}

	hsImage(hsImage* iLU, hsImage* iRD, int axis)
	{
		if(iLU->hasF || iRD->hasF)
			hasF = true;
		if(iLU->hasB || iRD->hasB)
			hasB = true;
		if(iLU->hasU || iLU->hasU)
			hasU = true;
		uX0 = (iLU->uX0 < iRD->uX0) ? iLU->uX0 : iRD->uX0;
		uY0 = (iLU->uY0 < iRD->uY0) ? iLU->uY0 : iRD->uY0;
		uX1 = (iLU->uX1 > iRD->uX1) ? iLU->uX1 : iRD->uX1;
		uY1 = (iLU->uY1 > iRD->uY1) ? iLU->uY1 : iRD->uY1;
		AkStar = NULL;	BkStar = NULL;
		meanPixel = NULL;	covarianceMatrix = NULL;

		//concatenate in x axis
		if(axis == 1) 
		{
			width = iLU->width + iRD->width;
			height = iLU->height;
			nChannels = iLU->nChannels;
			numPixels = width*height;
		}
		else //concatenate in y axis
		{
			width = iLU->width;
			height = iLU->height + iRD->height;
			nChannels = iLU->nChannels;
			numPixels = width * height;
		}

		iPixel = new float[width*height*nChannels];
		constraint = new int[width*height];
		alphaMap = new float[width*height];
		vX = new float[width*height];
		vP = new float[width*height];
		vR = new float[width*height];

		if(axis==1) //X-axis concatenate
		{
			int i=0;
			for(int h=0; h<height; h++)
			{
				for(int lw=0; lw<iLU->width; lw++)
				{
					for(int c=0; c<nChannels; c++)
						iPixel[i*nChannels + c] = iLU->iPixel[h*(iLU->width*nChannels) + lw*nChannels + c];
					constraint[i] = iLU->constraint[h*iLU->width + lw];
					alphaMap[i] = iLU->alphaMap[h*iLU->width + lw];
					vX[i] = iLU->vX[h*iLU->width + lw];
					vP[i] = iLU->vP[h*iLU->width + lw];
					vR[i] = iLU->vR[h*iLU->width + lw];
					i++;
				}
				for(int rw=0; rw<iRD->width; rw++)
				{
					for(int c=0; c<nChannels; c++)
						iPixel[i*nChannels + c] = iRD->iPixel[h*(iLU->width*nChannels) + rw*nChannels + c];
					constraint[i] = iLU->constraint[h*iLU->width + rw];
					alphaMap[i] = iLU->alphaMap[h*iLU->width + rw];
					vX[i] = iRD->vX[h*iRD->width + rw];
					vP[i] = iRD->vP[h*iRD->width + rw];
					vR[i] = iRD->vR[h*iRD->width + rw];
					i++;
				}
			}
		}
		else //Y-axis concatenate
		{
			for(int i=0; i<iLU->numPixels; i++)
			{
				for(int c=0; c<nChannels; c++)
					iPixel[i*nChannels +  c] = iLU->iPixel[i*nChannels +  c];
				constraint[i] = iLU->constraint[i];
				alphaMap[i] = iLU->alphaMap[i];
				vX[i] = iLU->vX[i];
				vP[i] = iLU->vP[i];
				vR[i] = iLU->vR[i];
			}
			int di = iLU->numPixels;
			for(int i=0; i<iRD->numPixels; i++)
			{
				for(int c=0; c<nChannels; c++)
					iPixel[di*nChannels +  c] = iRD->iPixel[i*nChannels +  c];
				constraint[di] = iRD->constraint[i];
				alphaMap[di] = iRD->alphaMap[i];
				vX[di] = iRD->vX[i];
				vP[di] = iRD->vP[i];
				vR[di] = iRD->vR[i];
				di++;
			}
		}
	}

	void checkFBU()
	{
		for(int i=0; i<numPixels; i++)
			if(constraint[i] == 1 && alphaMap[i] > 0.99f)
				hasF = true;
			else if(constraint[i] == 1 && alphaMap[i] < 0.01f)
				hasB = true;
			else if(constraint[i] == 0)
			{
				int x = i % width;
				int y = i / width;
				if(x < uX0)
					uX0 = x;
				if(y < uY0)
					uY0 = y;
				if(x > uX1)
					uX1 = x;
				if(x > uY1)
					uY1 = y;
				hasU = true;
			}
	}

	void calculateMeanPixel(int radius)
	{
		//If space was not allocated, allocate it
		if(meanPixel == NULL)
		{
			meanPixel = new float*[numPixels];
			for(int i=0; i<numPixels; i++)
				meanPixel[i] = new float[3];
		}
		#pragma omp parallel for
		for(int x=0; x<width; x++){
			for(int y=0; y<height; y++){
				int x0 = (x-radius > 0) ? x-radius : 0;
				int y0 = (y-radius > 0) ? y-radius : 0;
				int x1 = (x+radius < width) ? x+radius : width;
				int y1 = (y+radius < height) ? y+radius : height;
				int _w_ = (x1-x0) * (y1-y0);

				meanPixel[y*width + x][0] = 0.0f;
				meanPixel[y*width + x][1] = 0.0f;
				meanPixel[y*width + x][2] = 0.0f;
				for(int j=y0; j<y1; j++){
					for(int i=x0; i<x1; i++)
					{
						meanPixel[y*width + x][0] += iPixel[j*(width*nChannels) + i*nChannels + 0];
						meanPixel[y*width + x][1] += iPixel[j*(width*nChannels) + i*nChannels + 1];
						meanPixel[y*width + x][2] += iPixel[j*(width*nChannels) + i*nChannels + 2];
					}
				}
				meanPixel[y*width + x][0] /= _w_;
				meanPixel[y*width + x][1] /= _w_;
				meanPixel[y*width + x][2] /= _w_;
			}
		}
	}

	void calculateCovarianceMatrices(int radius)
	{
		//Allocate space if space is not allocated
		if(covarianceMatrix ==NULL)
		{
			covarianceMatrix = new float*[numPixels];
			for(int i=0; i<numPixels; i++){
				covarianceMatrix[i] = new float[9];
			}
		}
		if(meanPixel == NULL)
			calculateMeanPixel(radius);

		#pragma omp parallel for
		for(int x=0; x<width; x++){
			for(int y=0; y<height; y++){
				int x0 = (x-radius >= 0) ? x-radius : 0;
				int y0 = (y-radius >= 0) ? y-radius : 0;
				int x1 = (x+radius <= width) ? x+radius : width;
				int y1 = (y+radius <= height) ? y+radius : height;
				int _w_ = (x1-x0) * (y1-y0);
				int _i = y*width + x;

				covarianceMatrix[_i][0] = 0.0f; covarianceMatrix[_i][1] = 0.0f; covarianceMatrix[_i][2] = 0.0f;
				covarianceMatrix[_i][3] = 0.0f; covarianceMatrix[_i][4] = 0.0f; covarianceMatrix[_i][5] = 0.0f;
				covarianceMatrix[_i][6] = 0.0f; covarianceMatrix[_i][7] = 0.0f; covarianceMatrix[_i][8] = 0.0f;

				for(int j=y0; j<y1; j++){
					for(int i=x0; i<x1; i++){
						float cP[3] = {iPixel[j*(width*nChannels)+i*nChannels+0],
										iPixel[j*(width*nChannels)+i*nChannels+1],
										iPixel[j*(width*nChannels)+i*nChannels+2]};
						cP[0] -= meanPixel[_i][0];
						cP[1] -= meanPixel[_i][1];
						cP[2] -= meanPixel[_i][2];

						covarianceMatrix[_i][0] += cP[0]*cP[0]; covarianceMatrix[_i][1] += cP[0]*cP[1]; covarianceMatrix[_i][2] += cP[0]*cP[2];
						covarianceMatrix[_i][3] += cP[1]*cP[0]; covarianceMatrix[_i][4] += cP[1]*cP[1]; covarianceMatrix[_i][5] += cP[1]*cP[2];
						covarianceMatrix[_i][6] += cP[2]*cP[0]; covarianceMatrix[_i][7] += cP[2]*cP[1]; covarianceMatrix[_i][8] += cP[2]*cP[2];
					}
				}

				covarianceMatrix[_i][0]/=(_w_-1); covarianceMatrix[_i][1]/=(_w_-1); covarianceMatrix[_i][2]/=(_w_-1);
				covarianceMatrix[_i][3]/=(_w_-1); covarianceMatrix[_i][4]/=(_w_-1); covarianceMatrix[_i][5]/=(_w_-1);
				covarianceMatrix[_i][6]/=(_w_-1); covarianceMatrix[_i][7]/=(_w_-1); covarianceMatrix[_i][8]/=(_w_-1);
			}
		}
	}

	void calculateAkstarBkstar(int radius, float* p)
	{
		//Allocate space if space is not allocated
		if(AkStar == NULL)
		{
			AkStar = new float*[numPixels];
			for(int i=0; i<numPixels; i++){
				AkStar[i] = new float[3];
			}
		}
		if(BkStar == NULL)
		{
			BkStar = new float[numPixels];
		}

		float IDM[9] = {1,0,0,
			            0,1,0,
						0,0,1};

		for(int y=0; y<height; y++){
			for(int x=0; x<width; x++){
				int x0 = (x-radius >= 0) ? x-radius : 0;
				int y0 = (y-radius >= 0) ? y-radius : 0;
				int x1 = (x+radius <= width) ? x+radius : width;
				int y1 = (y+radius <= height) ? y+radius : height;
				int _w_ = (x1-x0) * (y1-y0);
				int _i = y*width + x;

				//get inverse of delta_k
				float delta_k[9];
				for(int i=0; i<9; i++)
					delta_k[i] = covarianceMatrix[_i][i] + IDM[i]*(_E_/_w_);
				CvMat Dk = cvMat(3,3,CV_32FC1,delta_k);
				CvMat* iDk = cvCreateMat(3,3,CV_32FC1);
				cvInvert(&Dk,iDk);
				delta_k[0] = cvmGet(iDk,0,0); delta_k[1] = cvmGet(iDk,0,1); delta_k[2] = cvmGet(iDk,0,2);
				delta_k[3] = cvmGet(iDk,1,0); delta_k[4] = cvmGet(iDk,1,1); delta_k[5] = cvmGet(iDk,1,2);
				delta_k[6] = cvmGet(iDk,2,0); delta_k[7] = cvmGet(iDk,2,1); delta_k[8] = cvmGet(iDk,2,2);

				//release iDk
				cvReleaseMat(&iDk);

				//divide by _w_;
				delta_k[0] /=_w_; delta_k[1] /=_w_; delta_k[2] /=_w_;
				delta_k[3] /=_w_; delta_k[4] /=_w_; delta_k[5] /=_w_;
				delta_k[6] /=_w_; delta_k[7] /=_w_; delta_k[8] /=_w_;

				//Get mean p
				float mean_p = 0.0f;
				for(int j=y0; j<y1; j++)
					for(int i=x0; i<x1; i++)
						mean_p += p[j*width + i];
				mean_p /= _w_;
				//Get mean_pixel * mean_p
				float mp_mp[3] = {mean_p*meanPixel[_i][0],mean_p*meanPixel[_i][1],mean_p*meanPixel[_i][2]};
				float Ii_pi[3] = {0.0f, 0.0f, 0.0f};
				float half_result[3] = {0.0f,0.0f,0.0f};

				for(int j=y0; j<y1; j++){
					for(int i=x0; i<x1; i++){
						Ii_pi[0] = iPixel[j*(width*nChannels)+i*nChannels+0] * p[j*width+i];
						Ii_pi[1] = iPixel[j*(width*nChannels)+i*nChannels+1] * p[j*width+i];
						Ii_pi[2] = iPixel[j*(width*nChannels)+i*nChannels+2] * p[j*width+i];
						Ii_pi[0] -= mp_mp[0];
						Ii_pi[1] -= mp_mp[1];
						Ii_pi[2] -= mp_mp[2];
						half_result[0] += Ii_pi[0];
						half_result[1] += Ii_pi[1];
						half_result[2] += Ii_pi[2];
					}
				}

				//Finally, get akstar using inverted delta and half result
				AkStar[_i][0] = delta_k[0] * half_result[0] + delta_k[1] * half_result[1] + delta_k[2] * half_result[2];
				AkStar[_i][1] = delta_k[3] * half_result[0] + delta_k[4] * half_result[1] + delta_k[5] * half_result[2];
				AkStar[_i][2] = delta_k[6] * half_result[0] + delta_k[7] * half_result[1] + delta_k[8] * half_result[2];

				//Now we know AkStar, so calculate BkStar
				BkStar[_i] = mean_p - (AkStar[_i][0] * meanPixel[_i][0] + AkStar[_i][1] * meanPixel[_i][1] + 
																	AkStar[_i][2] * meanPixel[_i][2]);
			}
		}
	}

	float* getMeanPixel(int i) {return meanPixel[i];}
	float* getCovarianceMatrix(int i) {return covarianceMatrix[i];}

	void writeXtoAlpha()
	{
		for(int i=0; i<numPixels; i++)
			alphaMap[i] = vX[i];
	}

	void deleteMeanPixel()
	{
		for(int i=0; i<numPixels; i++)
			delete[] meanPixel[i];
		delete[] meanPixel;
		meanPixel = NULL;
	}
	void deleteCovarianceMatrix()
	{
		for(int i=0; i<numPixels; i++)
			delete[] covarianceMatrix[i];
		delete[] covarianceMatrix;
		covarianceMatrix = NULL;
	}
	void deleteAkStar()
	{
		for(int i=0; i<numPixels; i++)
			delete[] AkStar[i];
		delete[] AkStar;
		AkStar = NULL;
	}
	void deleteBkStar()
	{
		delete[] BkStar;
		BkStar = NULL;
	}

	~hsImage()
	{
		delete[] iPixel;
		delete[] constraint;
		delete[] alphaMap;
		delete[] vX;
		delete[] vP;
		delete[] vR;

		if(meanPixel != NULL)
			deleteMeanPixel();
		if(covarianceMatrix != NULL)
			deleteCovarianceMatrix();
		if(AkStar != NULL)
			deleteAkStar();
		if(BkStar != NULL)
			deleteBkStar();
	}
};